
//#ifndef BIGUINT

// define the default hamming distance function
#define bui_fasthamming bui_fasthammingx64 // this function is very efficient for 64 bit calculations
//#define bui_fasthamming bui_fasthamming_builtin // if we need L > 64 use this one

#ifdef __cplusplus
extern "C" {
#endif

// By using new datatypes for input and output data, our program is able to read and
// process any kind of input and output datatypes, this approach keeps the code clean
// and the required changes minimal

//*

// 1: unsigned char, 4: float, 8: double
#ifndef BA_INTYPE
#define BA_INTYPE 8
#endif
#ifndef BA_OUTTYPE
#define BA_OUTTYPE 8
#endif

// define the input type and it's conversion to double functions
#if BA_INTYPE == 1
typedef unsigned char intype;
#define intype2double(x) ((double)(x) / 255.0)
#define double2intype(x) ((intype)(x * 255.0 + 0.5))
#elif BA_INTYPE == 4
typedef float intype;
#define intype2double(x) ((double)(x))
#define double2intype(x) ((intype)(x))
//#elif BA_INTYPE == 8
#else
typedef double intype;
#define intype2double(x) ((double)(x))
#define double2intype(x) ((intype)(x))
#endif

// define the output type and it's conversion to double functions
#if BA_OUTTYPE == 1
typedef unsigned char outype;
#define outype2double(x) ((double)(x) / 255.0)
#define double2outype(x) ((outype)(x * 255.0 + 0.5))
#elif BA_OUTTYPE == 4
typedef float outype;
#define outype2double(x) ((double)(x))
#define double2outype(x) ((outype)(x))
//#elif BA_OUTTYPE == 8
#else
typedef double outype;
#define outype2double(x) ((double)(x))
#define double2outype(x) ((outype)(x))
#endif

//*/

/*
// define the input type and it's conversion to double functions
typedef unsigned char intype;
#define intype2double(x) ((double)(x) / 255.0)
#define double2intype(x) ((intype)(x * 255.0 + 0.5))

// define the output type and it's conversion to double functions
typedef unsigned char outype;
#define outype2double(x) ((double)(x) / 255.0)
#define double2outype(x) ((outype)(x * 255.0 + 0.5))
//*/

// in order to be able to apply efficient bit operations on binary numbers of any size
// we have defined this new struct which represents a big unsigned int, also the required
// functions are implemented for this new datatype

#define bui_blen 2 // the total number of available bits is bui_blen * sizeof(unsigned int)
typedef struct
{
	unsigned int b[bui_blen];
}biguint;

void bui_zero(biguint *bui);
void bui_setbit(biguint *bui, int bit, int val);
int bui_fasthamming_builtin(biguint *bui1, biguint *bui2);
int bui_fasthammingx64(biguint *bui1, biguint *bui2);
int bui_equal(biguint *bui1, biguint *bui2); // 0:bui1!=bui2 1:bui1==bui2
void bui_inc(biguint *bui);
int bui_getbit(biguint *bui, int bit);
void bui_set(biguint *bui1, biguint *bui2);

#ifdef __cplusplus
}
#endif

//#endif
//#define BIGUINT
