#include "BA.h"

// deepnetforward takes input X and neural net F, return the output fX.
void deepnetforward(gsl_matrix* X, deepnet_layer* F[], int Nlayers, gsl_matrix* fX, gsl_matrix* Y, double* fiterr)
{
	int N = X->size1;
	int Din = X->size2;

	// T1 stores the input to each layer.
	gsl_matrix* T1 = gsl_matrix_calloc(N, Din);
	gsl_matrix_memcpy(T1, X);

	int i, j;
	for(i = 0; i < Nlayers; i++)
	{
		int Dout = F[i]->units;
		gsl_matrix* temp = gsl_matrix_calloc(N, Dout);

		layerforward(T1, F[i], temp, NULL);

		gsl_matrix_free(T1);
		T1 = temp;
	}

	// Finally copy the output.
	gsl_matrix_memcpy(fX, T1);
	gsl_matrix_free(T1);

	if (Y)
	{
		int Dout = Y->size2;
		*fiterr = 0;

		if (F[Nlayers - 1]->type == LAYER_TYPE_SOFTMAX)
		{
			for(i = 0; i < N; i++)
				for(j = 0; j < Dout; j++)
				{
					if (gsl_matrix_get(fX, i, j) == 0)
					{
						fprintf(stdout, "error\n");
					}
					(*fiterr) += -gsl_matrix_get(Y, i, j) * log(gsl_matrix_get(fX, i, j));
				}
		}
		else
		{
			for(i = 0; i < N; i++)
				for(j = 0; j < Dout; j++)
				{
					double t = gsl_matrix_get(Y, i, j) - gsl_matrix_get(fX, i, j);
					(*fiterr) += t * t;
				}
		}
	}
}

void layerforward(gsl_matrix* X, deepnet_layer* f, gsl_matrix* fX, gsl_matrix* J)
{
	switch(f->type)
	{
	case LAYER_TYPE_LINEAR: // Linear Layer
		linearforward(X, f, fX, J);
		break;
	case LAYER_TYPE_SIGMOID: // Sigmoid Layer
		//sigmoidforward(X, f, fX, J);
		break;
	case LAYER_TYPE_SOFTMAX: // Softmax Layer
		//softmaxforward(X, f, fX, J);
		break;
	}
}

void layererror(gsl_matrix* Y, deepnet_layer* f, gsl_matrix* fX, double* fiterr, double* regerr)
{
	int i, j;
	int N = Y->size1;
	int Din = f->W->size1 - 1;
	int Dout = Y->size2;
	double t;

	*fiterr = 0;
	if (f->type == LAYER_TYPE_SOFTMAX)
	{
		for(i = 0; i < N; i++)
			for(j = 0; j < Dout; j++)
				(*fiterr) += -gsl_matrix_get(Y, i, j) * log(gsl_matrix_get(fX, i, j));
	}
	else
	{
		// Compute sum squared error.
		for(i = 0; i < N; i++)
			for(j = 0; j < Dout; j++)
			{
				t = gsl_matrix_get(Y, i, j) - gsl_matrix_get(fX, i, j);
				(*fiterr) += t * t;
			}
	}

	*regerr = 0;
	// Compute regularization (for weights only).
	for(i = 0; i < Din; i++)
		for(j = 0; j < Dout; j++)
		{
			t = gsl_matrix_get(f->W, i, j);
			(*regerr) += f->l * t * t;
		}
}
