//============================================================================
// Name        : fstep.c
// Author      :
// Version     :
// Copyright   : Copyright (c) 2014 by Mehdi Alizadeh
// Description :
//============================================================================

#include "BA.h"

deepnet_layer f;
gsl_matrix *Wtemp;

int breaklayer;
deepnet_layer fslice;

void Finit(int L, int Dx, int breakf)
{
	f.l = 0;
	f.W = gsl_matrix_alloc(L + 1, Dx);
	f.type = LAYER_TYPE_LINEAR;
	f.units = Dx;
	gsl_matrix_set_all(f.W, 0.5);

	Wtemp = gsl_matrix_alloc(L + 1, /*1*/Dx);

	breaklayer = breakf;
	fslice.l = f.l;
	fslice.W = gsl_matrix_alloc(f.W->size1, 1);
	fslice.type = f.type;
	f.units = 1;
}

void Ftrain(gsl_matrix *Z, gsl_matrix *X)
{
	lineartrain(Z, X, &f, NULL, NULL, NULL);
}

void Fforward(gsl_matrix *X, gsl_matrix *fX)
{
	linearforward(X, &f, fX, NULL);
}

void Ferror(biguint *Z, outype *X, long int N, int Dx, int L, double *Ferr)
{
	gsl_matrix *Ztemp = gsl_matrix_alloc(1, L);
	gsl_matrix *Xtemp = gsl_matrix_alloc(1, Dx);
	gsl_matrix *fZtemp = gsl_matrix_alloc(1, Dx);

	double subFerr;
	*Ferr = 0;

	long int i;
	int j;
	for (i = 0; i < N; i++)
	{
		for (j = 0; j < L; j++)
			Ztemp->data[j] = bui_getbit(&(Z[i]), j);
		for (j = 0; j < Dx; j++)
			Xtemp->data[j] = outype2double(X[i * Dx + j]);

		double notused;
		linearforward(Ztemp, &f, fZtemp, NULL);
		layererror(Xtemp, &f, fZtemp, &subFerr, &notused);
		*Ferr += subFerr;
	}

	gsl_matrix_free(Ztemp);
	gsl_matrix_free(Xtemp);
	gsl_matrix_free(fZtemp);
}

void FtrainCG(biguint *Z, outype *X, long int N, int Dx, int L, int chunks, int epochs)
{
	int chunksize = N / chunks + 1;
	gsl_matrix *Ztemp = gsl_matrix_alloc(chunksize, L);
	gsl_matrix *Xtemp = gsl_matrix_alloc(chunksize, Dx);

	int ichunk, iepoch;
	for (iepoch = 0; iepoch < epochs; iepoch++)
	{
		long int ibegin = 0;

		for (ichunk = 0; ichunk < chunks; ichunk++)
		{
			long int ilen = (N - ibegin) / (chunks - ichunk);

			if (ilen == 0)
				continue;

			if (ilen != Ztemp->size1)
			{
				gsl_matrix_free(Ztemp);
				gsl_matrix_free(Xtemp);
				Ztemp = gsl_matrix_alloc(ilen, L);
				Xtemp = gsl_matrix_alloc(ilen, Dx);
			}

			int i, j;
			for (i = 0; i < ilen; i++)
			{
				for (j = 0; j < L; j++)
					Ztemp->data[i * L + j] = bui_getbit(&(Z[i + ibegin]), j);
				for (j = 0; j < Dx; j++)
					Xtemp->data[i * Dx + j] = outype2double(X[(i + ibegin) * Dx + j]);
			}

			lineartrainCG(Ztemp, Xtemp, &f, 2, NULL, NULL, NULL);

			ibegin += ilen;
		}
	}

	gsl_matrix_free(Ztemp);
	gsl_matrix_free(Xtemp);
}

void FtrainSGD(biguint *Z, outype *X, long int N, int Dx, int L, int chunks, int epochs, int iann, int stepcounter, int *minibatchperm)
{
	long int chunksize = N / chunks + 1;
	gsl_matrix *Ztemp = gsl_matrix_alloc(chunksize, L);
	gsl_matrix *Xtemp = gsl_matrix_alloc(chunksize, breaklayer ? 1 : Dx);

	int counter = stepcounter * chunks;

	int i, ichunk, iepoch;

	long int begins[chunks];
	long int lens[chunks];

	long int ibegin = 0;
	for (ichunk = 0; ichunk < chunks; ichunk++)
	{
		long int ilen = (N - ibegin) / (chunks - ichunk);

		begins[ichunk] = ibegin;
		lens[ichunk] = ilen;

		ibegin += ilen;
	}

	deepnet_layer *fp;
	if (breaklayer)
	{
		int temp = 0;
		getFmodel(iann, fslice.W->data, &temp);
		fp = &fslice;
	}
	else
		fp = &f;

	for (iepoch = 0; iepoch < epochs; iepoch++)
	{
		for (i = 0; i < chunks; i++)
		{
			ichunk = minibatchperm[i];

			if (lens[ichunk] == 0)
				continue;

			if (lens[ichunk] != Ztemp->size1)
			{
				gsl_matrix_free(Ztemp);
				gsl_matrix_free(Xtemp);
				Ztemp = gsl_matrix_alloc(lens[ichunk], L);
				Xtemp = gsl_matrix_alloc(lens[ichunk], breaklayer ? 1 : Dx);
			}

			long int p;
			int q;
			for (p = 0; p < lens[ichunk]; p++)
				for (q = 0; q < L; q++)
					Ztemp->data[p * L + q] = bui_getbit(&(Z[p + begins[ichunk]]), q);

			if (breaklayer)
			{
				for (p = 0; p < lens[ichunk]; p++)
					Xtemp->data[p] = outype2double(X[(p + begins[ichunk]) * Dx + iann]);
			}
			else
			{
				for (p = 0; p < lens[ichunk]; p++)
					for (q = 0; q < Dx; q++)
						Xtemp->data[p * Dx + q] = outype2double(X[(p + begins[ichunk]) * Dx + q]);
			}

			double k = 0.0001;
			double step = k / (k * counter + 1.0);
			lineartrainSD(Ztemp, Xtemp, fp, step, NULL, NULL, NULL);
			counter += 1;
		}
	}

	if (breaklayer)
	{
		int temp = 0;
		setFmodel(iann, fslice.W->data, &temp);
	}

	gsl_matrix_free(Ztemp);
	gsl_matrix_free(Xtemp);
}

void FtrainexactSGD(gsl_matrix *Z, gsl_matrix *X, int chunks, int epochs, double alpha, int iann, int reset)
{
	int N = Z->size1;
	int Dx = X->size2;
	int L = Z->size2;

	int ichunk, iepoch;
	for (iepoch = 0; iepoch < epochs; iepoch++)
	{
		int ibegin = 0;

		for (ichunk = 0; ichunk < chunks; ichunk++)
		{
			int ilen = (N - ibegin) / (chunks - ichunk);

			if (ilen == 0)
				continue;

			gsl_matrix_view Zview = gsl_matrix_submatrix(Z, ibegin, 0, ilen, L);
			gsl_matrix_view Xview = gsl_matrix_submatrix(X, ibegin, 0, ilen, Dx);

			gsl_matrix_memcpy(Wtemp, f.W);
			lineartrain(&(Zview.matrix), &(Xview.matrix), &f, NULL, NULL, NULL);
			//lineartrainCG(&(Zview.matrix), &(Xview.matrix), &f, 5, NULL, NULL, NULL);

			gsl_matrix_scale(Wtemp, 1.0 - alpha);
			gsl_matrix_scale(f.W, alpha);
			gsl_matrix_add(f.W, Wtemp);

			double err;
			linearerror(&(Zview.matrix), &(Xview.matrix), &f, &err, NULL);

			ibegin += ilen;
		}
	}
}

void getFmodel(int iann, double *buffer, int *index)
{
	if (breaklayer)
	{
		int i, datasize = f.W->size1, Dx = f.W->size2;
		for (i = 0; i < datasize; i++)
			buffer[*index + i] = f.W->data[i * Dx + iann];
		*index += datasize;
	}
	else
	{
		int datasize = f.W->size1 * f.W->size2;
		memcpy(&(buffer[*index]), f.W->data, datasize * sizeof(double));
		*index += datasize;
	}
}

void setFmodel(int iann, double *buffer, int *index)
{
	if (breaklayer)
	{
		int i, datasize = f.W->size1, Dx = f.W->size2;
		for (i = 0; i < datasize; i++)
			f.W->data[i * Dx + iann] = buffer[*index + i];
		*index += datasize;
	}
	else
	{
		int datasize = f.W->size1 * f.W->size2;
		memcpy(f.W->data, &(buffer[*index]), datasize * sizeof(double));
		*index += datasize;
	}
}

