//============================================================================
// Name        : precision.c
// Author      :
// Version     :
// Copyright   : Copyright (c) 2014 by Mehdi Alizadeh
// Description :
//============================================================================

#include "BA.h"

double calcprecision(configstruct *config, int loadbinaries)
{
	int totalN = config->N;
	int testN = config->Ntest;
	int Din = config->Din;
	int L = config->L;

	static intype *trainset = NULL, *testset = NULL/*, *trainlabel = NULL, *testlabel = NULL*/;
	static biguint *trainbins, *testbins;

	static int *distancecount, *distances, *sortedindices;
	static unsigned int *testneighborindices;

	if (!trainset)
	{
		trainset = (intype*)malloc(totalN * Din  * sizeof(intype));
		freadbuffer(config->trainsetRBF, (unsigned char*)trainset, totalN * Din * sizeof(intype), 0);

		testset = (intype*)malloc(testN * Din * sizeof(intype));
		freadbuffer(config->testset, (unsigned char*)testset, testN * Din * sizeof(intype), 0);

		/*
		trainlabel = gsl_matrix_alloc(totalN, 1);
		freadmatrix(config->trainlabels, trainlabel);

		testlabel = gsl_matrix_alloc(testN, 1);
		freadmatrix(config->testlabels, testlabel);
	    */

		trainbins = (biguint*)calloc(totalN, sizeof(biguint));
		testbins = (biguint*)calloc(testN, sizeof(biguint));

		distancecount = (int*)malloc((L + 1) * sizeof(int));
		distances = (int*)malloc(totalN * sizeof(int));
		sortedindices = (int*)malloc(totalN * sizeof(int));

		int maxneighbors = config->K;
		testneighborindices = (unsigned int*)malloc(testN * maxneighbors * sizeof(unsigned int));
		freadbuffer(config->neighbors, (unsigned char*)(testneighborindices), testN * maxneighbors * sizeof(unsigned int), 0);
	}

	int trueneighborcount = config->K;
	int latentneighborcount = config->k;

	int correctneighbors = 0;
	int i, j, k;

	if (loadbinaries)
	{
		// load previously calculated binary hash values
		unsigned char *bins = (unsigned char*)malloc(totalN * L);
		freadbuffer(config->initbintrain, bins, totalN * L, 0);
		for (j = 0; j < totalN * L; j++)
			bui_setbit(&(trainbins[j / L]), j % L, bins[j]);
		free(bins);

		bins = (unsigned char*)malloc(testN * L);
		freadbuffer(config->initbintest, bins, testN * L, 0);
		for (j = 0; j < testN * L; j++)
			bui_setbit(&(testbins[j / L]), j % L, bins[j]);
		free(bins);
	}
	else
	{
		// calculate train and test binary hash values;
		HforwardSGD(trainset, trainbins, totalN, L, Din);
		HforwardSGD(testset, testbins, testN, L, Din);
	}

	unsigned int cachmem[trueneighborcount];
	for (i = 0; i < testN; i++)
	{
		for (j = 0; j <= L; j++)
			distancecount[j] = 0;

		for (j = 0; j < totalN; j++)
		{
			distances[j] = bui_fasthamming(&(testbins[i]), &(trainbins[j]));
			distancecount[distances[j]]++;
		}

		for (j = 1; j <= L; j++)
			distancecount[j] += distancecount[j - 1];

		for (j = 0; j < totalN; j++)
		{
			int index = distancecount[distances[j]] - 1;
			distancecount[distances[j]]--;

			sortedindices[index] = j;
		}

		memcpy(cachmem, &(testneighborindices[i * config->K]), trueneighborcount * sizeof(unsigned int));

		for (j = 0; j < trueneighborcount; j++)
		{
			int index = (int)cachmem[j];
			for (k = 0; k < latentneighborcount; k++)
				if (index == sortedindices[k])
				{
					correctneighbors++;
					break;
				}
		}
	}

	double precision = (double)correctneighbors / testN / latentneighborcount;
	printf("precision: %f\n", precision);

	return precision;
}
