#include "BA.h"
#include <time.h>

void Ftest()
{
	int N = 4;
	int Din = 3;
	int Dout = 2;
	// we have 3x2 full layer

	double x[] = {0.46, 0.21, 0.30,
			0.53, 0.66, 0.79,
			0.39, 0.94, 0.13,
			0.82, 0.81, 0.28};
	gsl_matrix_view xv = gsl_matrix_view_array(x, N, Din);
	gsl_matrix* X = &(xv.matrix);

	double y[] = {0.67, 0.24,
			0.91, 0.69,
			0.18, 0.61,
			0.65, 0.10};
	gsl_matrix_view yv = gsl_matrix_view_array(y, N, Dout);
	gsl_matrix* Y = &(yv.matrix);

	double w[] = {0.1, 0.2,
			0.9, 0.3,
			0.3, 0.6,
			0.8, 0.7};
	gsl_matrix_view wv = gsl_matrix_view_array(w, Din + 1, Dout);
	gsl_matrix* W = &(wv.matrix);

	deepnet_layer* F = (deepnet_layer*)malloc(sizeof(deepnet_layer));
	F->l = 0.1;
	F->units = Dout;
	F->W = gsl_matrix_alloc(Din + 1, Dout);

	gsl_matrix* fX = gsl_matrix_alloc(N, Dout);
	gsl_matrix* J = gsl_matrix_alloc(Dout, Din);

	double fiterr, regerr;
	int maxit = 20;

	//linearlayer
	F->type = LAYER_TYPE_LINEAR;
	gsl_matrix_memcpy(F->W, W);

	fprintf(stdout, "\nlinearlayer forward:\n");
	linearforward(X, F, fX, J);
	print_matrix("fX", fX);
	print_matrix("jacobian", J);

	//lineartrain exact
	fprintf(stdout, "\nlinearlayer train:\n");
	lineartrain(X, Y, F, fX, &fiterr, &regerr);
	print_matrix("W", F->W);
	print_matrix("fX", fX);
	fprintf(stdout, "fiterr: %12.8f\n", fiterr);
	fprintf(stdout, "regerr: %12.8f\n", regerr);

	//lineartrainCG bylayer
	fprintf(stdout, "\nlinearlayer trainCG bylayer:\n");
	gsl_matrix_memcpy(F->W, W);

	lineartrainCG(X, Y, F, maxit, fX, &fiterr, &regerr);
	print_matrix("W", F->W);
	print_matrix("fX", fX);
	fprintf(stdout, "fiterr: %12.8f\n", fiterr);
	fprintf(stdout, "regerr: %12.8f\n", regerr);

	//lineartrainCG byunit
	fprintf(stdout, "\nlinearlayer trainCG byunit:\n");
	gsl_matrix_memcpy(F->W, W);

	lineartrainCG_byunit(X, Y, F, maxit, fX, &fiterr, &regerr);
	print_matrix("W", F->W);
	print_matrix("fX", fX);
	fprintf(stdout, "fiterr: %12.8f\n", fiterr);
	fprintf(stdout, "regerr: %12.8f\n", regerr);

	//lineartrainCG byunit
	fprintf(stdout, "\nlinearlayer trainSGD:\n");
	gsl_matrix_memcpy(F->W, W);

	int i;
	for (i = 0; i < maxit; i++)
	{
		lineartrainSD(X, Y, F, 1.0 / (i + 1), fX, &fiterr, &regerr);
		//fprintf(stdout, "iteration %d fiterr: %12.8f\n", i, fiterr);
	}

	print_matrix("W", F->W);
	print_matrix("fX", fX);
	fprintf(stdout, "fiterr: %12.8f\n", fiterr);
	fprintf(stdout, "regerr: %12.8f\n", regerr);
}

void Ztest()
{
	struct timeval begin, end;
	int i, j;
	double diff, time;

	int N = 5000;
	int D = 30;
	int L = 16;
	double mu = 0.0234;
	int block = 1;

	Finit(L, D, 0);
	Zinit(N, D, L);

	gsl_matrix* _X = gsl_matrix_alloc(N, D);
	gsl_matrix* _hX = gsl_matrix_alloc(N, L);
	gsl_matrix* _Z0 = gsl_matrix_alloc(N, L);
	gsl_matrix* _Z = gsl_matrix_alloc(N, L);
	gsl_matrix* _Zrlx = gsl_matrix_alloc(N, L);
	gsl_matrix* _Zltr = gsl_matrix_alloc(N, L);
	gsl_matrix* _myZ = gsl_matrix_alloc(N, L);

	freadmatrix("Ztest/ZtestX.bin", _X);
	freadmatrix("Ztest/ZtesthX.bin", _hX);
	freadmatrix("Ztest/ZtestZ0.bin", _Z0);
	freadmatrix("Ztest/ZtestZ.bin", _Z);
	freadmatrix("Ztest/ZtestZrelaxed.bin", _Zrlx);
	freadmatrix("Ztest/ZtestZalternate.bin", _Zltr);

	outype* X = (outype*)malloc(N * D * sizeof(outype));
	biguint* hX = (biguint*)malloc(sizeof(biguint) * N);
	biguint* Z0 = (biguint*)malloc(sizeof(biguint) * N);
	//biguint* Z = (biguint*)malloc(sizeof(biguint) * N);
	biguint* Zrlx = (biguint*)malloc(sizeof(biguint) * N);
	biguint* Zltr = (biguint*)malloc(sizeof(biguint) * N);
	biguint* myZ = (biguint*)malloc(sizeof(biguint) * N);

	for (i = 0; i < N; i++)
	{
		for(j = 0; j < D; j++)
			X[i * D + j] = double2outype(_X->data[i * D + j]);

		bui_zero(&(hX[i]));
		bui_zero(&(Z0[i]));
		bui_zero(&(Zrlx[i]));
		bui_zero(&(Zltr[i]));
		bui_zero(&(myZ[i]));

		for (j = 0; j < L; j++)
		{
			bui_setbit(&(hX[i]), j, (int)(_hX->data[i * L + j] + 0.1));
			bui_setbit(&(Z0[i]), j, (int)(_Z0->data[i * L + j] + 0.1));
			bui_setbit(&(Zrlx[i]), j, (int)(_Zrlx->data[i * L + j] + 0.1));
			bui_setbit(&(Zltr[i]), j, (int)(_Zltr->data[i * L + j] + 0.1));
			bui_setbit(&(myZ[i]), j, (int)(_myZ->data[i * L + j] + 0.1));
		}
	}

	Ftrain(_Z0, _X);

	gsl_matrix_memcpy(_myZ, _Z0);
	gettimeofday(&begin, NULL);
	//Zstepexact(_X, _hX, _myZ, mu);
	gettimeofday(&end, NULL);
	time = HighResDiff(end, begin);

	diff = 0;
	for (i = 0; i < N * L; i++)
			diff += fabs(_Z->data[i] - _myZ->data[i]);

	printf("exact Zstep\ndiff: %12.8f time: %12.8f\n", diff, time);

	gsl_matrix_memcpy(_myZ, _Z0);
	gettimeofday(&begin, NULL);
	//Zstepfastexact(_X, _hX, _myZ, mu);
	gettimeofday(&end, NULL);
	time = HighResDiff(end, begin);

	diff = 0;
	for (i = 0; i < N * L; i++)
		diff += fabs(_Z->data[i] - _myZ->data[i]);

	printf("fastexact Zstep\ndiff: %12.8f time: %12.8f\n", diff, time);


	Zinit(1, D, L);

	memcpy(myZ, Z0, sizeof(biguint) * N);
	gettimeofday(&begin, NULL);
	ZstepalternateBUI(X, hX, myZ, N, D, L, mu, block, 0);
	gettimeofday(&end, NULL);
	time = HighResDiff(end, begin);

	Herror(Zltr, myZ, &diff, N, L);

	printf("alternative Zstep\ndiff: %12.8f time: %12.8f\n", diff, time);

	memcpy(myZ, Z0, sizeof(biguint) * N);
	gettimeofday(&begin, NULL);
	ZsteprelaxedBUI(X, hX, myZ, N, D, L, mu);
	ZstepalternateBUI(X, hX, myZ, N, D, L, mu, 1, 0);
	gettimeofday(&end, NULL);
	time = HighResDiff(end, begin);

	Herror(Zrlx, myZ, &diff, N, L);

	printf("relaxed Zstep\ndiff: %12.8f time: %12.8f\n", diff, time);

}

void fastKNN(int datapoints, int bits)
{
	int i, j;
	struct timeval begin, end;
	double t;

	srand(time(NULL));
	biguint *points = (biguint*)malloc(datapoints * sizeof(biguint));
	for (i = 0; i < datapoints; i++)
		for (j = 0; j < 2; j++)
			points[i].b[j] = (unsigned int)(rand() * rand());

	int dist = 0;
	gettimeofday(&begin, NULL);
	for (i = 1; i < datapoints; i++)
		dist += bui_fasthamming(&(points[0]), &(points[i]));
	gettimeofday(&end, NULL);
	t = HighResDiff(end, begin);
	printf("KNN search time: %6.2fs est: %4.1fhrs dist: %d\n",
			t * (1000000000 / datapoints), t * (1000000000 / datapoints) * 10000 / 3600, dist);

	dist = 0;
	gettimeofday(&begin, NULL);
	for (i = 1; i < datapoints; i++)
		dist += bui_fasthammingx64(&(points[0]), &(points[i]));
	gettimeofday(&end, NULL);
	t = HighResDiff(end, begin);
	printf("KNN search time: %6.2fs est: %4.1fhrs dist: %d\n",
			t * (1000000000 / datapoints), t * (1000000000 / datapoints) * 10000 / 3600, dist);

}













