//============================================================================
// Name        : zstep.c
// Author      :
// Version     :
// Copyright   : Copyright (c) 2014 by Mehdi Alizadeh
// Description :
//============================================================================

#include "BA.h"

biguint binlimit;
gsl_matrix *Zdomain, *fZdomain;
double *Xslice;

gsl_matrix *A; // Dx*L
gsl_vector *tau; // L
gsl_matrix *Q, *R; // Dx*Dx, Dx*L
gsl_matrix *rZ; // 2^L*L
gsl_matrix *XQ; // N*L
gsl_matrix *Xtemp; // N*Dx
double *XQslice; // L

gsl_matrix *Rview, *Qview;

gsl_matrix *_A;// LxL
gsl_matrix *_X1;// NxD
gsl_matrix *_b;// NxL
gsl_matrix *_svdU; // LxL
gsl_matrix *_svdV; // LxL
gsl_vector *_svdS; // L
gsl_vector *_svdWork; // L

gsl_matrix *_mV; // NxL
gsl_matrix *_Y; // NxL
gsl_matrix *_X; //NxL
gsl_matrix *_oldZ; // NxL

gsl_matrix *mX;
gsl_matrix *mhX;
gsl_matrix *mZ;

extern deepnet_layer f;

void Zinit(long int N, int Dx, int L)
{
	int bit;
	bui_zero(&binlimit);
	for (bit = 0; bit < L; bit++)
		bui_setbit(&binlimit, bit, 1);

	Zdomain = gsl_matrix_alloc(1, L);
	fZdomain = gsl_matrix_alloc(1, Dx);

	Xslice = (double*)malloc(Dx * sizeof(double));

	A = gsl_matrix_alloc(Dx, L);
	tau = gsl_vector_alloc(L);
	Q = gsl_matrix_alloc(Dx, Dx);
	R = gsl_matrix_alloc(Dx, L);
	rZ = gsl_matrix_alloc(1, L);
	XQ = gsl_matrix_alloc(N, L);
	Xtemp = gsl_matrix_alloc(N, Dx);
	XQslice = (double*)malloc(L * sizeof(double));

	Rview = gsl_matrix_alloc(L, L);
	Qview = gsl_matrix_alloc(Dx, L);

	_A = gsl_matrix_alloc(L, L);// LxL
	_X1 = gsl_matrix_alloc(N, Dx);// NxD
	_b = gsl_matrix_alloc(N, L);// NxL
	_svdU = gsl_matrix_alloc(L, L); // LxL
	_svdV = gsl_matrix_alloc(L, L); // LxL
	_svdS = gsl_vector_alloc(L); // L
	_svdWork = gsl_vector_alloc(L); // L

	_mV = gsl_matrix_alloc(N, L); // NxL
	_Y = gsl_matrix_alloc(N, L); // NxL
	_X = gsl_matrix_alloc(N, L); //NxL
	_oldZ = gsl_matrix_alloc(N, L); // NxL

	mX = gsl_matrix_alloc(N, Dx);
	mhX = gsl_matrix_alloc(N, L);
	mZ = gsl_matrix_alloc(N, L);
}

void Zstepexact(gsl_matrix *X, gsl_matrix *hX, gsl_matrix *Z, double mu)
{
	int N = X->size1;
	int Dx = X->size2;
	int L = hX->size2;

	int i, j, bit;
	biguint z;

	for (i = 0; i < N; i++)
	{
		double mincost = -1;
		biguint mincostZ, hx;
		bui_zero(&mincostZ);
		bui_zero(&hx);

		for (bit = 0; bit < L; bit++)
			if ((int)gsl_matrix_get(hX, i, bit))
				bui_setbit(&hx, bit, 1);

		for (j = 0; j < Dx; j++)
			Xslice[j] = gsl_matrix_get(X, i, j);

		for (bui_zero(&z); ; bui_inc(&z))
		{
			double cost = mu * bui_fasthamming(&z, &hx);

			for (bit = 0; bit < L; bit++)
				Zdomain->data[bit] = bui_getbit(&z, bit);
			Fforward(Zdomain, fZdomain);

			for (j = 0; j < Dx; j++)
			{
				double temp = fZdomain->data[j] - Xslice[j];
				cost += temp * temp;
			}

			if (cost < mincost || mincost < 0)
			{
				mincost = cost;
				bui_set(&mincostZ, &z);
			}

			if (bui_equal(&z, &binlimit) == 1)
				break;
		}

		for (bit = 0; bit < L; bit++)
			gsl_matrix_set(Z, i, bit, bui_getbit(&mincostZ, bit));
	}
}

void Zstepfastexact(gsl_matrix *X, gsl_matrix *hX, gsl_matrix *Z, double mu)
{
	int N = X->size1;
	int Dx = X->size2;
	int L = hX->size2;

	int i, j, bit, p, q;

	gsl_matrix_view Wview = gsl_matrix_submatrix(f.W, 0, 0, L, Dx);
	gsl_matrix_transpose_memcpy(A, &(Wview.matrix));
	gsl_linalg_QR_decomp(A, tau);
	gsl_linalg_QR_unpack(A, tau, Q, R);

	gsl_matrix_view _Rview = gsl_matrix_submatrix(R, 0, 0, L, L);
	gsl_matrix_view _Qview = gsl_matrix_submatrix(Q, 0, 0, Dx, L);
	gsl_matrix_memcpy(Rview, &(_Rview.matrix));
	gsl_matrix_memcpy(Qview, &(_Qview.matrix));

	gsl_matrix_memcpy(Xtemp, X);
	gsl_matrix_view bias = gsl_matrix_submatrix(f.W, L, 0, 1, Dx);
	for (i = 0; i < N; i++)
	{
		gsl_matrix_view Xtempview = gsl_matrix_submatrix(Xtemp, i, 0, 1, Dx);
		gsl_matrix_sub(&(Xtempview.matrix), &(bias.matrix));
	}
	gsl_blas_dgemm (CblasNoTrans, CblasNoTrans, 1.0, Xtemp, Qview, 0.0, XQ);

	biguint z;
	double cachmem[L];
	for (i = 0; i < N; i++)
	{
		double mincost = -1;
		biguint mincostZ, hx;
		bui_zero(&mincostZ);
		bui_zero(&hx);

		memcpy(cachmem, &(hX->data[i * L]), L * sizeof(double));

		for (bit = 0; bit < L; bit++)
			if ((int)cachmem[bit]) //gsl_matrix_get(hX, i, bit)
				bui_setbit(&hx, bit, 1);

		memcpy(XQslice, &(XQ->data[i * L]), L * sizeof(double));

		for (bui_zero(&z); ; bui_inc(&z))
		{
			double cost = mu * bui_fasthamming(&z, &hx);

			for (bit = 0; bit < L; bit++)
				Zdomain->data[bit] = bui_getbit(&z, bit);
			//gsl_blas_dgemm (CblasNoTrans, CblasTrans, 1.0, Zdomain, Rview, 0.0, rZ);
			for (p = 0; p < L; p++)
			{
				rZ->data[p] = 0;
				for (q = 0; q < L; q++)
					rZ->data[p] += Zdomain->data[q] * Rview->data[p * L + q];
			}

			for (j = 0; j < L; j++)
			{
				double temp = rZ->data[j] - XQslice[j]; // gsl_matrix_get(rZ, z, j)
				cost += temp * temp;
			}

			if (cost < mincost || mincost < 0)
			{
				mincost = cost;
				bui_set(&mincostZ, &z);
			}

			if (bui_equal(&z, &binlimit) == 1)
				break;
		}

		for (bit = 0; bit < L; bit++)
			gsl_matrix_set(Z, i, bit, bui_getbit(&mincostZ, bit));
	}
}

inline void Zstepalternate_1(outype *X, biguint *hX, biguint *Z, long int N, int Dx, int L, double mu, int maxit)
{
	long int i;
	int j, bit, block, p, q, blockz;

	//gsl_matrix_memcpy(Xtemp, X);
	for (i = 0; i < N * Dx; i++)
		Xtemp->data[i] = outype2double(X[i]);

	gsl_matrix_view bias = gsl_matrix_submatrix(f.W, L, 0, 1, Dx);
	for (i = 0; i < N; i++)
	{
		gsl_matrix_view Xtempview = gsl_matrix_submatrix(Xtemp, i, 0, 1, Dx);
		gsl_matrix_sub(&(Xtempview.matrix), &(bias.matrix));
	}
	gsl_blas_dgemm (CblasNoTrans, CblasNoTrans, 1.0, Xtemp, Qview, 0.0, XQ);

	for (i = 0; i < N; i++)
	{
		memcpy(XQslice, &(XQ->data[i * L]), L * sizeof(double));

		double hXerr = 0;
		for (bit = 0; bit < L; bit++)
			Zdomain->data[bit] = bui_getbit(&(hX[i]), bit);
		for (p = 0; p < L; p++)
		{
			rZ->data[p] = 0;
			for (q = 0; q < L; q++)
				rZ->data[p] += Zdomain->data[q] * Rview->data[p * L + q];
		}
		for (j = 0; j < L; j++)
		{
			double temp = rZ->data[j] - XQslice[j]; //gsl_matrix_get(rZ, z, j)
			hXerr += temp * temp;
		}
		if (hXerr < mu)
		{
			bui_set(&(Z[i]), &(hX[i]));
			continue;
		}

		double mincost = -1;
		biguint mincostZ, hx, z, zalt;

		bui_zero(&mincostZ);
		bui_set(&hx, &(hX[i]));
		bui_set(&z, &(Z[i]));

		int itcount = 0;
		while (maxit == 0 || itcount < maxit)
		{
			int flag = 0;
			for (block = 0; block < L; block++)
			{
				for (blockz = 0; blockz <= 1; blockz++)
				{
					//zalt = z & (~blockmask);
					//zalt |= blockz << (blocksize * block);
					bui_set(&zalt, &z);
					bui_setbit(&zalt, block, blockz);

					double cost = mu * bui_fasthamming(&zalt, &hx);

					for (bit = 0; bit < L; bit++)
						Zdomain->data[bit] = bui_getbit(&zalt, bit);
					//gsl_blas_dgemm (CblasNoTrans, CblasTrans, 1.0, Zdomain, Rview, 0.0, rZ);
					for (p = 0; p < L; p++)
					{
						rZ->data[p] = 0;
						for (q = 0; q < L; q++)
							rZ->data[p] += Zdomain->data[q] * Rview->data[p * L + q];
					}

					for (j = 0; j < L; j++)
					{
						double temp = rZ->data[j] - XQslice[j]; //gsl_matrix_get(rZ, z, j)
						cost += temp * temp;
					}

					if (cost < mincost || mincost < 0)
					{
						mincost = cost;
						bui_set(&mincostZ, &zalt);
					}
				}

				if (bui_equal(&mincostZ, &z) == 0)
				{
					bui_set(&z, &mincostZ);
					flag = 1;
				}
			}

			itcount++;
			if (flag == 0)
				break;
		}

		bui_set(&(Z[i]), &mincostZ);
	}
}

inline void Zstepalternate(outype *X, biguint *hX, biguint *Z, long int N, int Dx, int L, double mu, int blocksize, int maxit)
{
	if (blocksize == 1)
	{
		Zstepalternate_1(X, hX, Z, N, Dx, L, mu, maxit);
		return;
	}

	long int i;
	int j, k, bit, block, p, q;

	biguint blocklimit;
	bui_zero(&blocklimit);
	for (bit = 0; bit < blocksize; bit++)
		bui_setbit(&blocklimit, bit, 1);
	int blockcount = L / blocksize;

	//gsl_matrix_memcpy(Xtemp, X);
	for (i = 0; i < N * Dx; i++)
		Xtemp->data[i] = outype2double(X[i]);

	gsl_matrix_view bias = gsl_matrix_submatrix(f.W, L, 0, 1, Dx);
	for (i = 0; i < N; i++)
	{
		gsl_matrix_view Xtempview = gsl_matrix_submatrix(Xtemp, i, 0, 1, Dx);
		gsl_matrix_sub(&(Xtempview.matrix), &(bias.matrix));
	}
	gsl_blas_dgemm (CblasNoTrans, CblasNoTrans, 1.0, Xtemp, Qview, 0.0, XQ);

	biguint blockz, blockmask;
	for (i = 0; i < N; i++)
	{
		memcpy(XQslice, &(XQ->data[i * L]), L * sizeof(double));

		double hXerr = 0;
		for (bit = 0; bit < L; bit++)
			Zdomain->data[bit] = bui_getbit(&(hX[i]), bit);
		for (p = 0; p < L; p++)
		{
			rZ->data[p] = 0;
			for (q = 0; q < L; q++)
				rZ->data[p] += Zdomain->data[q] * Rview->data[p * L + q];
		}
		for (j = 0; j < L; j++)
		{
			double temp = rZ->data[j] - XQslice[j]; //gsl_matrix_get(rZ, z, j)
			hXerr += temp * temp;
		}
		if (hXerr < mu)
		{
			bui_set(&(Z[i]), &(hX[i]));
			continue;
		}

		double mincost = -1;
		biguint mincostZ, hx, z, zalt;

		bui_zero(&mincostZ);
		bui_set(&hx, &(hX[i]));
		bui_set(&z, &(Z[i]));

		int itcount = 0;
		while (maxit == 0 || itcount < maxit)
		{
			int flag = 0;
			for (block = 0; block < blockcount; block++)
			{
				bui_zero(&blockmask);
				for (k = 0; k < blocksize; k++)
					bui_setbit(&blockmask, blocksize * block + k, 1);
				//blockmask = blocklimit << (blocksize * block);

				for (bui_zero(&blockz); ; bui_inc(&blockz))
				{
					//zalt = z & (~blockmask);
					//zalt |= blockz << (blocksize * block);
					bui_set(&zalt, &z);
					for (k = 0; k < blocksize; k++)
						bui_setbit(&zalt, blocksize * block + k, bui_getbit(&blockz, k));

					double cost = mu * bui_fasthamming(&zalt, &hx);

					for (bit = 0; bit < L; bit++)
						Zdomain->data[bit] = bui_getbit(&zalt, bit);
					//gsl_blas_dgemm (CblasNoTrans, CblasTrans, 1.0, Zdomain, Rview, 0.0, rZ);
					for (p = 0; p < L; p++)
					{
						rZ->data[p] = 0;
						for (q = 0; q < L; q++)
							rZ->data[p] += Zdomain->data[q] * Rview->data[p * L + q];
					}

					for (j = 0; j < L; j++)
					{
						double temp = rZ->data[j] - XQslice[j]; //gsl_matrix_get(rZ, z, j)
						cost += temp * temp;
					}

					if (cost < mincost || mincost < 0)
					{
						mincost = cost;
						bui_set(&mincostZ, &zalt);
					}

					if (bui_equal(&blockz, &blocklimit) == 1)
						break;
				}

				if (bui_equal(&mincostZ, &z) == 0)
				{
					bui_set(&z, &mincostZ);
					flag = 1;
				}
			}

			itcount++;
			if (flag == 0)
				break;
		}

		bui_set(&(Z[i]), &mincostZ);
	}
}

void proxbqp(gsl_matrix *V, double m, gsl_matrix *cholA, double r, gsl_matrix *b, gsl_matrix *Z)
{
	int N = V->size1;
	int L = V->size2;

	int maxit = 100;
	double tol = 1e-3;

	gsl_matrix_memcpy(_mV, V);
	gsl_matrix_scale(_mV, m);

	gsl_matrix_set_zero(_Y);

	int i, j;
	for (i = 0; i < maxit; i++)
	{
		gsl_matrix_memcpy(_oldZ, Z);

		gsl_matrix_sub(Z, _Y);
		gsl_matrix_scale(Z, r);
		gsl_matrix_add(Z, b);

		for (j = 0; j < N; j++)
		{
			gsl_vector_view vb = gsl_vector_view_array(&(Z->data[j * L]), L);
			gsl_vector_view vx = gsl_vector_view_array(&(_X->data[j * L]), L);

			gsl_linalg_cholesky_solve(cholA, &(vb.vector), &(vx.vector));
		}

		gsl_matrix_memcpy(Z, _X);
		gsl_matrix_add(Z, _Y);
		gsl_matrix_scale(Z, r);
		gsl_matrix_add(Z, _mV);
		gsl_matrix_scale(Z, 1.0 / (m + r));

		for (j = 0; j < N * L; j++)
			if (Z->data[j] < 0)
				Z->data[j] = 0;
			else if (Z->data[j] > 1)
				Z->data[j] = 1;

		gsl_matrix_add(_Y, _X);
		gsl_matrix_sub(_Y, Z);

		double diff, maxdiff = -1;
		for (j = 0; j < N * L; j++)
		{
			diff = fabs(Z->data[j] - _oldZ->data[j]);
			if (diff > maxdiff)
				maxdiff = diff;
		}
		if (maxdiff < tol)
			break;
	}
}

void Zsteprelaxed(gsl_matrix *X, gsl_matrix *hX, gsl_matrix *Z, double mu, double r)
{
	int i;

	int N = X->size1;
	int Dx = X->size2;
	int L = hX->size2;

	gsl_matrix_memcpy(_X1, X);
	for (i = 0; i < N * Dx; i++)
		_X1->data[i] -= f.W->data[L * Dx + (i % Dx)];

	gsl_matrix_view Wview = gsl_matrix_submatrix(f.W, 0, 0, L, Dx);
	gsl_blas_dgemm (CblasNoTrans, CblasTrans, 1.0, _X1, &(Wview.matrix), 0.0, _b);

	proxbqp(hX, mu, _A, r, _b, Z);

	for (i = 0; i < N * L; i++)
		Z->data[i] = (Z->data[i] > 0.5 ? 1 : 0);
}

void ZsteprelaxedBUI(outype *X, biguint *hX, biguint *Z, long int N, int Dx, int L, double mu)
{
	long int i;
	int j;

	double r = 0;
	gsl_matrix_view Wview = gsl_matrix_submatrix(f.W, 0, 0, L, Dx);

	gsl_blas_dgemm (CblasNoTrans, CblasTrans, 1.0, &(Wview.matrix), &(Wview.matrix), 0.0, _A);

	gsl_matrix_memcpy(_svdU, _A);
	gsl_linalg_SV_decomp(_svdU, _svdV, _svdS, _svdWork);
	r = sqrt(_svdS->data[0] * _svdS->data[L - 1]);

	for (i = 0; i < L; i++)
		_A->data[i * L + i] += r;
	gsl_linalg_cholesky_decomp(_A);

	for (i = 0; i < N; i++)
	{
		for (j = 0; j < Dx; j++)
			mX->data[j] = outype2double(X[i * Dx + j]);

		for (j = 0; j < L; j++)
		{
			mhX->data[j] = bui_getbit(&(hX[i]), j);
			mZ->data[j] = bui_getbit(&(Z[i]), j);
		}

		Zsteprelaxed(mX, mhX, mZ, mu, r);
		for (j = 0; j < L; j++)
			bui_setbit(&(Z[i]), j, (int)(mZ->data[j]));
	}
}

void ZstepalternateBUI(outype *X, biguint *hX, biguint *Z, long int N, int Dx, int L, double mu, int blocksize, int maxit)
{
	gsl_matrix_view Wview = gsl_matrix_submatrix(f.W, 0, 0, L, Dx);
	gsl_matrix_transpose_memcpy(A, &(Wview.matrix));
	gsl_linalg_QR_decomp(A, tau);
	gsl_linalg_QR_unpack(A, tau, Q, R);

	gsl_matrix_view _Rview = gsl_matrix_submatrix(R, 0, 0, L, L);
	gsl_matrix_view _Qview = gsl_matrix_submatrix(Q, 0, 0, Dx, L);
	gsl_matrix_memcpy(Rview, &(_Rview.matrix));
	gsl_matrix_memcpy(Qview, &(_Qview.matrix));

	long int i;
	for (i = 0; i < N; i++)
		Zstepalternate(&(X[i * Dx]), &(hX[i]), &(Z[i]), 1, Dx, L, mu, blocksize, maxit);
}

